#!/usr/bin/env python
#============================================================================#
#                                                                            #
# VendorID; SIB (Signed Interpreter Builder)                                 #
#                                                                            #
# Copyright (c) 2005 DESYS GmbH; All Rights Reserved                         #
#                                                                            #
# This file is part of VendorID.                                             #
#                                                                            #
# This copy of VendorID is licensed for use under the terms and              #
# conditions of the VendorID License Agreement. See the file                 #
# LICENSE.txt for more details.                                              #
#                                                                            #
# VendorID is supplied WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.           #
#                                                                            #
#----------------------------------------------------------------------------#
# $Id: sib.py,v 1.1.1.1 2005/01/07 10:03:58 berning Exp $
#----------------------------------------------------------------------------#
import sys, os, marshal, string, getopt, glob, re

#============================================================================#
# Use this icon for windows executables if no other icon is given            #
#----------------------------------------------------------------------------#
windef_icon = ("\x00\x00\x01\x00\x01\x00\x20\x20\x10\x00\x00\x00\x00\x00\xe8\x02"
               "\x00\x00\x16\x00\x00\x00\x28\x00\x00\x00\x20\x00\x00\x00\x40\x00"
               "\x00\x00\x01\x00\x04\x00\x00\x00\x00\x00\x80\x02\x00\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x00\x00\x00\x80\x00\x00\x80\x00\x00\x00\x80\x80\x00\x80\x00"
               "\x00\x00\x80\x00\x80\x00\x80\x80\x00\x00\x80\x80\x80\x00\xc0\xc0"
               "\xc0\x00\x00\x00\xff\x00\x00\xff\x00\x00\x00\xff\xff\x00\xff\x00"
               "\x00\x00\xff\x00\xff\x00\xff\xff\x00\x00\xff\xff\xff\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x00\xaa\xa3\xa0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x03\x33\xaa\xaa\xa0\x0a\xaa\xaa\x00\x00\x00\x00\x00\x00\x00"
               "\x00\xaa\xaa\xa3\x33\x33\xa3\x33\x33\x33\x00\x00\x00\x00\x00\x00"
               "\x00\xaa\xaa\x33\x33\x33\x33\x33\x33\x33\x30\x0a\xa0\x00\x00\x00"
               "\x00\x33\x3a\x33\x33\x33\x33\x33\x33\x33\x33\x33\x33\x00\x00\x00"
               "\x00\xaa\xaa\x33\x33\x30\x33\x33\x33\x03\x33\x33\x33\x30\x00\x00"
               "\x00\x0a\xaa\x33\x33\x33\x03\x33\x30\x00\x33\x30\x03\x30\x00\x00"
               "\x00\x03\x33\xa3\x33\x33\x30\x00\x00\x00\x00\x00\x00\x30\x00\x90"
               "\x00\x00\x0a\xaa\x33\x33\x33\x00\x00\x00\x00\x00\x00\x00\x90\x90"
               "\x00\x00\x00\x33\xaa\x33\x33\x30\x00\x00\x00\x00\x00\x00\x09\x90"
               "\x00\x00\x00\x00\xaa\xa3\x33\x33\x00\x00\x00\x00\x00\x00\x00\x99"
               "\xaa\xaa\x00\x00\x03\x3a\x33\x33\x30\x00\x00\x00\x00\x00\x00\x09"
               "\x9a\x3a\xaa\x00\x00\x0a\xa3\x33\x33\x00\x00\x00\x00\x00\x00\x00"
               "\x99\x00\x3a\xaa\x00\x00\x3a\x33\x33\x00\x00\x00\x00\x00\x00\x00"
               "\x09\x90\x00\x3a\xa0\x00\xaa\xa3\x33\x30\x00\x00\x00\x00\x00\xa3"
               "\xa3\xa3\x00\x0a\x3a\x00\x33\xa3\x33\x30\x00\x00\x00\x00\x0a\xaa"
               "\xaa\xaa\xa0\x00\x3a\xa0\xaa\xa3\x33\x30\x00\x00\x00\x00\x0a\xa3"
               "\xa3\xa3\xaa\xa0\x0a\x3a\x0a\x33\x33\x30\x00\x00\x00\x00\x03\x33"
               "\x33\x33\xa3\xaa\x00\xaa\xa0\x33\x33\x30\x00\x00\x00\x00\x03\x03"
               "\x30\x33\x33\xa3\xaa\xaa\xa0\x33\x33\x00\x00\x00\x00\x00\x03\x33"
               "\x33\x33\x33\x33\xa3\xaa\xa0\x33\x33\x00\x00\x00\x00\x00\x00\x33"
               "\x33\x33\x33\x33\x33\x3a\x03\x33\x30\x00\x00\x00\x00\x00\x00\x00"
               "\x33\x33\x33\x33\x33\x30\x33\x30\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x00\x30\x00\x33\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\xbb\x0b\xbb\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\xbc\x0b\xcc\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\xbc\x0b\xcc\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\xbb\x0b\xbb\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x00\x00\x00\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x03\x30\x33\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x07"
               "\xff\xff\xfe\x01\x83\xff\xfc\x00\x00\xff\xf8\x00\x00\x67\xf8\x00"
               "\x00\x03\xf8\x00\x00\x01\xf8\x00\x00\x00\xfc\x00\x02\x00\xfc\x00"
               "\x07\x18\xde\x00\x1f\xfc\x5f\x80\x0f\xfe\x80\xe0\x07\xff\xc0\x30"
               "\x03\xff\xc0\x0c\x01\xff\xc0\x06\x01\xff\xc0\x02\x00\xff\x80\x00"
               "\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00\x00\xff\x00\x00"
               "\x01\xff\x00\x00\x01\xff\x80\x00\x03\xff\xc0\x00\x07\xff\xf0\x00"
               "\x1f\xff\xf8\x00\xff\xff\xf8\x01\xff\xff\xf8\x01\xff\xff\xf8\x01"
               "\xff\xff\xf8\x01\xff\xff\xfc\x03\xff\xff\xfe\x47\xff\xff")

#============================================================================#
# Use this icon for console executables if no other icon is given            #
#----------------------------------------------------------------------------#
condef_icon = ("\x00\x00\x01\x00\x01\x00\x20\x20\x10\x00\x00\x00\x00\x00\xe8\x02"
               "\x00\x00\x16\x00\x00\x00\x28\x00\x00\x00\x20\x00\x00\x00\x40\x00"
               "\x00\x00\x01\x00\x04\x00\x00\x00\x00\x00\x80\x02\x00\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x00\x00\x00\x80\x00\x00\x80\x00\x00\x00\x80\x80\x00\x80\x00"
               "\x00\x00\x80\x00\x80\x00\x80\x80\x00\x00\x80\x80\x80\x00\xc0\xc0"
               "\xc0\x00\x00\x00\xff\x00\x00\xff\x00\x00\x00\xff\xff\x00\xff\x00"
               "\x00\x00\xff\x00\xff\x00\xff\xff\x00\x00\xff\xff\xff\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07"
               "\x77\x77\x77\x77\x77\x77\x77\x77\x77\x77\x77\x77\x70\x00\x00\xf8"
               "\x88\x88\x88\x88\x88\x88\x88\x88\x88\x88\x88\x88\x87\x00\x00\xf8"
               "\x8f\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xf8\x87\x00\x00\xf8"
               "\x70\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0f\x87\x00\x00\xf8"
               "\x70\x99\x00\x00\x00\x00\x0a\xaa\x33\x33\x30\x0f\x87\x00\x00\xf8"
               "\x70\x09\x9a\xaa\xa0\x00\x00\x33\xa3\x33\x33\x0f\x87\x00\x00\xf8"
               "\x70\x00\x99\xa3\xaa\xa0\x00\x00\xaa\x33\x33\x0f\x87\x00\x00\xf8"
               "\x70\x00\x09\x90\x03\xaa\xa0\x00\x03\xa3\x33\x0f\x87\x00\x00\xf8"
               "\x70\x00\x00\x99\x00\x03\xaa\x00\x0a\xaa\x33\x0f\x87\x00\x00\xf8"
               "\x70\x0a\x3a\x3a\x30\x00\xa3\xa0\x03\x3a\x33\x0f\x87\x00\x00\xf8"
               "\x70\xaa\xaa\xaa\xaa\x00\x03\xaa\x0a\xaa\x33\x0f\x87\x00\x00\xf8"
               "\x70\xaa\x3a\x3a\x3a\xaa\x00\xa3\xa0\xa3\x33\x0f\x87\x00\x00\xf8"
               "\x70\x33\x33\x33\x3a\x3a\xa0\x0a\xaa\x03\x33\x0f\x87\x00\x00\xf8"
               "\x70\x30\x33\x03\x33\x3a\x3a\xaa\xaa\x03\x33\x0f\x87\x00\x00\xf8"
               "\x70\x33\x33\x33\x33\x33\x3a\x3a\xaa\x03\x33\x0f\x87\x00\x00\xf8"
               "\x70\x03\x33\x33\x33\x33\x33\x33\xa0\x33\x33\x0f\x87\x00\x00\xf8"
               "\x70\x00\x03\x33\x33\x33\x33\x33\x03\x33\x00\x0f\x87\x00\x00\xf8"
               "\x70\x00\x00\x00\x03\x00\x03\x33\x00\x00\x00\x0f\x87\x00\x00\xf8"
               "\x70\x00\x00\x0b\xb0\xbb\xb0\x30\x00\x00\x00\x0f\x87\x00\x00\xf8"
               "\x70\x00\x00\x0b\xc0\xbc\xc0\x30\x00\x00\x00\x0f\x87\x00\x00\xf8"
               "\x70\x00\x00\x0b\xc0\xbc\xc0\x30\x00\x00\x00\x0f\x87\x00\x00\xf8"
               "\x70\x00\x00\x0b\xb0\xbb\xb0\x30\x00\x00\x00\x0f\x87\x00\x00\xf8"
               "\x70\x00\x00\x00\x00\x00\x03\x30\x00\x00\x00\x0f\x87\x00\x00\xf8"
               "\x70\x00\x00\x00\x33\x03\x33\x00\x00\x00\x00\x0f\x87\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xcc"
               "\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\x00\x00\x77"
               "\xcf\xfc\xff\xcf\xff\xcc\xfc\xcc\xcc\xcc\xc7\x7c\x77\x00\x00\x87"
               "\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xc8\x7c\x87\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
               "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff"
               "\xff\xff\xe0\x00\x00\x07\xc0\x00\x00\x03\x80\x00\x00\x01\x80\x00"
               "\x00\x01\x80\x00\x00\x01\x80\x07\x00\x01\x82\x01\x80\x01\x82\x00"
               "\x60\x01\x82\x00\x30\x01\x82\x00\x10\x01\x80\x00\x00\x01\x80\x00"
               "\x00\x01\x80\x00\x00\x01\x80\x00\x00\x01\x80\x00\x00\x01\x80\x00"
               "\x00\x01\x80\x00\x00\x01\x82\x00\x00\x01\x83\x80\x00\xc1\x83\xc0"
               "\x07\xc1\x83\xc0\x0f\xc1\x83\xc0\x0f\xc1\x83\xc0\x0f\xc1\x83\xc0"
               "\x0f\xc1\x83\xe0\x1f\xc1\x80\x00\x00\x01\x80\x00\x00\x01\x80\x00"
               "\x00\x01\x80\x00\x00\x01\x80\x00\x00\x01\xff\xff\xff\xff")

#============================================================================#
# This class holds all information about a frozen script/module.             #
#----------------------------------------------------------------------------#
class FrozenCode:
    def __init__(self, module_name, input_file, prefix):
        sys.stdout.write("Compiling '%s' using prefix '%s'\n" % (input_file, prefix))

        # Save the input file name
        self.input_file = input_file
        
        # Save the module name
        self.module_name = module_name

        # Create a valid C name by replacing '.' with '_'.
        self.mangled_name = module_name.replace('.', '_')

        # Open the input file and read the source.
        # Append a newline (compile() needs at least one)
        f = open(input_file, 'r')
        source_code = string.join([x.rstrip() for x in f.readlines()], '\n') + '\n'
        f.close()

        # Compile the code using prefix and marshall it.
        compiled_code = compile(source_code, prefix, 'exec')
        marshalled_code = marshal.dumps(compiled_code)

        # Store the size of the marshalled code
        self.size = len(marshalled_code)

        # Generate a character array
        l1 = []
        l2 = []
        i = 0
        for b in marshalled_code:
            l1.append('%d' % ord(b))
            i += 1
            if i == 16:
                l2.append(string.join(l1, ","))
                l1 = []
                i = 0
        if l1:
            l2.append(string.join(l1, ","))
        code_array = string.join(l2, ",\n    ")
        code_array += ","

        # Store the variable name of the character array
        self.c_name = 'M_' + self.mangled_name

        # Store the C code
        self.c_code = 'unsigned char %s[] =\n{\n    %s\n};\n\n' % (self.c_name, code_array)

#============================================================================#
# is_package() - Check, if a given package directory is a valid package      #
#----------------------------------------------------------------------------#
def is_package(package_dir):
    package_file = glob.glob(os.path.join(package_dir, '__init__.py*'))
    if package_file:
        return 1
    return 0

#============================================================================#
# resolve_package() - Walk down a given package directory and build the code #
#                     lists for every python module                          #
#----------------------------------------------------------------------------#
def resolve_package(package, parent, prefix, frozen_modules=[]):
    # Check, if the package is a valid python package
    if not is_package(package):
        return
    # Get the list of files/directories in the package directory
    package_files = os.listdir(package)
    # Collect subpackages and regular python files
    sub_packages = []
    package_modules = []
    for package_file in package_files:
        if package_file == '__init__.py':
            continue
        # Build the absolute path of the file/directory in the package
        full_name = os.path.join(package, package_file)
        # If the entry is a directory, check if it is a subpackage and collect it
        if os.path.isdir(full_name):
            if is_package(full_name):
                sub_packages.append(full_name)
        ext = os.path.splitext(package_file)[1]
        # Ignore any file non python files
        if not ext or ext != '.py':
            continue
        # Ignore any entries, that are not regular files
        if not os.path.isfile(full_name):
            continue
        # Collect the regular python file
        package_modules.append(full_name)
        
    # Process the package file (__init__.py)
    if prefix is None:
        p = os.path.join(package, '__init__.py')
    else:
        p = os.path.join(prefix, os.path.basename(package), '__init__.py')
    module_name = parent + os.path.splitext(os.path.basename(package))[0]
    fm = FrozenCode(module_name, os.path.join(package, '__init__.py'), p)
    # Package files have a negative size
    fm.size = -fm.size
    frozen_modules.append(fm)

    # Process every python module in the package directory
    for package_module in package_modules:
        if prefix is None:
            p = package_module
        else:
            p = os.path.join(prefix, os.path.basename(package), os.path.basename(package_module))
        module_name = (parent +
                       os.path.splitext(os.path.basename(package))[0] +
                       '.' +
                       os.path.splitext(os.path.basename(package_module))[0])
        frozen_modules.append(FrozenCode(module_name, package_module, p))

    # Process every subpackage in the package directory
    for sub_package in sub_packages:
        resolve_package(sub_package, (parent +
                                      os.path.splitext(os.path.basename(package))[0] +
                                      '.'), prefix, frozen_modules)
    
#============================================================================#
# Generate a resource file containing the applications icon (Windows only)   #
#----------------------------------------------------------------------------#
def generate_rc(output_dir, name, icon_name, console, sources):
    file_name = os.path.join(output_dir, '%s.rc' % name)
    if icon_name:
        sys.stdout.write("Generating recource file '%s' using icon '%s'\n" % (file_name, icon_name))    
    else:
        sys.stdout.write("Generating recource file '%s' using builtin icon\n" % file_name)    
    rc = open(file_name, 'w')
    rc.write('/*%s*/\n' % ('=' * 74))
    rc.write('/* Resource code for %s %s*/\n' % (name, (' ' * (54 - len(name)))))
    rc.write('/* This is generated code; Do not modify it!                                */\n')
    rc.write('/*%s*/\n' % ('-' * 74))
    rc.write('#define IDI_ICON 100\n')
    if not icon_name:
        ico_name = '%s.ico' % name
        f = open(os.path.join(output_dir, ico_name), 'wb')
        if console:
            f.write(condef_icon)
        else:
            f.write(windef_icon)
        f.close()
    else:
        ico_name = os.path.basename(icon_name)
        f = open(icon_name, 'rb')
        d = f.read()
        f.close()
        f = open(os.path.join(output_dir, ico_name), "wb")
        f.write(d)
        f.close()
    rc.write('IDI_ICON ICON DISCARDABLE "%s"\n\n' % ico_name)
    rc.write('/*%s*/\n' % ('=' * 74))
    rc.write('/* EOF%s*/\n' % (' ' * 70))
    rc.write('/*%s*/\n' % ('-' * 74))
    rc.close()
    sources.append(file_name)
    
#============================================================================#
# Generate the main C code                                                   #
#----------------------------------------------------------------------------#
def generate_main(output_dir, name, frozen_main, frozen_modules,
                  console, unsigned, optimize, verbose, sources, use_cxx):
    if use_cxx:
        file_name = os.path.join(output_dir, 'frozen.cc')
    else:
        file_name = os.path.join(output_dir, 'frozen.c')
    sys.stdout.write("Generating main startup code '%s'\n" % file_name)    
    f = open(file_name, 'w')
    f.write('/*%s*/\n' % ('=' * 74))
    f.write('/* Main startup code for %s %s*/\n' % (name, (' ' * (50 - len(name)))))
    f.write('/* This is generated code; Do not modify it!                                */\n')
    f.write('/*%s*/\n' % ('-' * 74))
    f.write('#include <stdarg.h>\n')
    f.write('#include <Python.h>\n')
    f.write('#include <marshal.h>\n')

    if sys.platform == 'win32' and not console:
        f.write('#include <windows.h>\n')

    if not unsigned:
        f.write('#include <vendorid.h>\n')

    f.write('\n')

    f.write('extern unsigned char %s[];\n' % frozen_main.c_name)

    for fm in frozen_modules:
        f.write('extern unsigned char %s[];\n' % fm.c_name)

    f.write('\n')
    
    f.write("static struct _frozen _PyImport_FrozenModules[] =\n{\n")
    for fm in frozen_modules:
        f.write('    {(char *)"%s", %s, %d},\n' % (fm.module_name, fm.c_name, fm.size))
    f.write('    {0, 0, 0} /* sentinel */\n};\n\n')

    if sys.platform == 'win32' and not console:
        f.write('int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,\n')
        f.write('                   LPSTR lpCmdLine, int nCmdShow)\n')
        f.write('{\n')
        f.write('    int argc = __argc;\n')
        f.write('    char **argv = __argv;\n')
    else:
        if use_cxx:
            f.write('extern "C" ')
        f.write('int main(int argc, char **argv)\n')
        f.write('{\n')
    f.write('    PyObject *v, *co;\n')
    if verbose:
        f.write('    register int i, j;\n')
    f.write('    PyImport_FrozenModules = _PyImport_FrozenModules;\n\n')
    if optimize > 0:
        f.write('    Py_OptimizeFlag++;\n')
    if optimize > 1:
        f.write('    Py_OptimizeFlag++;\n')
    if verbose:
        f.write('    for(i = 1; i < argc; i++)\n')
        f.write('    {\n')
        f.write('        if(!strcmp(argv[i], "-v") || !strcmp(argv[i], "-vv"))\n')
        f.write('        {\n')
        f.write('            Py_VerboseFlag++;\n')
        f.write('            if(!strcmp(argv[i], "-vv"))\n')
        f.write('                Py_VerboseFlag++;\n')
        f.write('            for(j = i; j < argc - 1; j++)\n')
        f.write('                argv[j] = argv[j + 1];\n')
        f.write('            argc--;\n')
        f.write('            break;\n')
        f.write('        }\n')
        f.write('    }\n')
    f.write('    Py_SetProgramName(argv[0]);\n')
    if not unsigned:        
        f.write('    vendorid_init();\n')
    f.write('    Py_Initialize();\n')
    f.write('    PySys_SetArgv(argc, argv);\n')
    f.write('    co = PyMarshal_ReadObjectFromString((char *)%s, %d);\n' % (frozen_main.c_name, frozen_main.size))
    f.write('    v = PyImport_ExecCodeModule((char *)"__main__", co);\n')
    f.write('    if(v == NULL)\n')
    f.write('    {\n')
    f.write('        PyErr_Print();\n')
    f.write('        return -1;\n')
    f.write('    }\n')
    f.write('    Py_DECREF(v);\n')
    f.write('    Py_Finalize();\n')
    f.write('    return 0;\n')
    f.write('}\n\n')
    f.write('/*%s*/\n' % ('=' * 74))
    f.write('/* EOF%s*/\n' % (' ' * 70))
    f.write('/*%s*/\n' % ('-' * 74))
    f.close()
    sources.append(file_name)

#============================================================================#
# Parse the default python makefile and collect all variables                #
#----------------------------------------------------------------------------#
def parse_unix_makefile():
    py_config_dir = os.path.join(sys.exec_prefix, 'lib', 'python%s' % sys.version[:3], 'config')
    variable_def = re.compile('^([a-zA-Z0-9_]+)[ \t]*=(.*)')
    variables = {}
    fp = open(os.path.join(py_config_dir, 'Makefile'), 'r')
    pendingline = ""
    try:
        while 1:
            line = fp.readline()
            if pendingline:
                    line = pendingline + line
                    pendingline = ""
            if not line:
                    break
            if line.endswith('\\\n'):
                    pendingline = line[:-2]
            matchobj = variable_def.match(line)
            if not matchobj:
                    continue
            (name, value) = matchobj.group(1, 2)
            # Strip trailing comment
            i = value.find('#')
            if i >= 0:
                    value = value[:i]
            value = value.strip()
            variables[name] = value
    finally:
        fp.close()
    return variables

#============================================================================#
# Expand Makefile variables in a string using a dictionary for the values    #
#----------------------------------------------------------------------------#
def expand_var(vars, value):
    # Check, if there is something to expand at all
    if '$' not in value:
        return value
    # We search for $(...) or ${...}
    pattern = re.compile(r'\$(\([^(]*\)|\{[^{]*\})')
    i = 0
    while 1:
        m = pattern.search(value, i)
        if not m:
            break
        i, j = m.span(0)
        name = m.group(1)
        if ((name.startswith('(') and name.endswith(')')) or
            (name.startswith('{') and name.endswith('}'))):
            name = name[1:-1]
        if vars.has_key(name):
            tail = value[j:]
            value = value[:i] + expand_var(vars, vars[name])
            i = len(value)
            value += tail
        else:
            i = j
    return value
    
#============================================================================#
# Generate a UNIX Makefile                                                   #
#----------------------------------------------------------------------------#
def unix_makefile(output_dir, exe_name, sources, makevars, unsigned, target_dir):
    file_name = os.path.join(output_dir, 'Makefile')
    sys.stdout.write("Generating %s'\n" % file_name)
    f = open(file_name, 'w')
    f.write('#%s#\n' % ('=' * 76))
    f.write('# Makefile for %s %s #\n' % (exe_name, (' ' * (60 - len(exe_name)))))
    f.write('# This is a generated file; Do not modify it!                                #\n')
    f.write('#%s#\n' % ('-' * 76))

    # The standard python directories.
    f.write('PYINCDIR = %s\n' % os.path.join(sys.prefix, 'include', 'python%s' % sys.version[:3]))
    f.write('PYCNFDIR = %s\n' % os.path.join(sys.exec_prefix, 'lib', 'python%s' % sys.version[:3], 'config'))
    f.write('PYLIBDIR = %s\n' % os.path.join(sys.exec_prefix, 'lib'))

    # Look, if we have the python library in <sys.exec_prefix>/lib.
    # This is the case when Python was configured with --enable-shared.
    shared_python_lib = glob.glob(os.path.join(sys.exec_prefix, 'lib', 'libpython%s.*' % sys.version[:3]))
    if shared_python_lib:
        libdir_name = '$(PYLIBDIR)'
    else:
        libdir_name = '$(PYCNFDIR)'

    # Append the python library to the LDLAST variable.
    if makevars.has_key('LDLAST'):
        makevars['LDLAST'] += ' -L%s' % libdir_name
        makevars['LDLAST'] += ' -lpython$(VERSION)'
    else:
        makevars['LDLAST'] = '-L%s' % libdir_name
        makevars['LDLAST'] += ' -lpython$(VERSION)'

    # A signed interpreter needs the vendorid library.
    if not unsigned:        
        if shared_python_lib:
            makevars['LDLAST'] += ' -L$(PYCNFDIR)'
        makevars['LDLAST'] += ' -lvendorid'

    # AIX needs some modifications
    if sys.platform.startswith('aix'):
        makevars['LINKCC'] = '$(CC)'
        makevars['LINKFORSHARED'] = makevars['LINKFORSHARED'].replace('Modules/python.exp', '$(PYCNFDIR)/python.exp')
    
    # Modify the cpp flags
    if makevars.has_key('CPPFLAGS'):
        makevars['CPPFLAGS'] = makevars['CPPFLAGS'].replace('-I$(srcdir)/Include', '-I$(PYINCDIR)')
    else:
        makevars['CPPFLAGS'] = '-I. -I$(PYINCDIR)'

    # Write all makefile variables (only those we need, but already expanded)
    for key in ('CC', 'CXX', 'CFLAGS', 'CPPFLAGS',
                'LINKCC', 'LDFLAGS', 'LINKFORSHARED',
                'LIBS', 'SYSLIBS', 'LDLAST'):
        if makevars.has_key(key):
            value = expand_var(makevars, makevars[key])
            f.write('%s = %s\n' % (key, value))
        
    # Generate and write the list of object files
    objs = []
    for source in sources:
        ext = os.path.splitext(source)[1]
        if ext in ('.c', '.cc', '.cpp', '.cxx'):
            objs.append(os.path.splitext(os.path.basename(source))[0] + '.o')
    prefix = 'FROZEN_OBJS = '
    for obj in objs:
        f.write('%s%s' % (prefix, obj))
        prefix = ' \\\n              '
    f.write('\n')

    # The target directory where to install
    f.write('TARGET_DIR = %s\n\n' % target_dir)

    # The target we want to create
    f.write('TARGET = %s\n\n' % exe_name)
        
    f.write('all: $(TARGET)\n\n')

    f.write('$(TARGET): $(FROZEN_OBJS)\n')
    f.write('\t$(LINKCC) $(LDFLAGS) $(LINKFORSHARED) $(FROZEN_OBJS) -o $(TARGET) $(LIBS) $(SYSLIBS) $(LDLAST)\n\n')

    # Generate the install target
    f.write('install: $(TARGET)\n')
    f.write('\t@test -d $(TARGET_DIR) || mkdir -p $(TARGET_DIR)\n')
    f.write('\tcp -f $(TARGET) $(TARGET_DIR)/$(TARGET)\n\n')

    # Generate the clean target
    f.write('clean:\n')
    f.write('\t-rm -f *.o $(TARGET)\n\n')

    # Generic rule for C sources
    f.write('.c.o:\n')
    f.write('\t$(CC) $(CFLAGS) $(CPPFLAGS) -c  $< -o $@\n\n')

    # Generic rule for C++ sources
    f.write('.cc.o:\n')
    f.write('\t$(CXX) $(CFLAGS) $(CPPFLAGS) -c  $< -o $@\n\n')
    f.write('.cpp.o:\n')
    f.write('\t$(CXX) $(CFLAGS) $(CPPFLAGS) -c  $< -o $@\n\n')
    f.write('.cxx.o:\n')
    f.write('\t$(CXX) $(CFLAGS) $(CPPFLAGS) -c  $< -o $@\n\n')
                
    f.write('#%s#\n' % ('=' * 76))
    f.write('# EOF%s#\n' % (' ' * 72))
    f.write('#%s#\n' % ('-' * 76))
    f.close()

#============================================================================#
# Generate a Windows Makefile                                                #
#----------------------------------------------------------------------------#
def win32_makefile(output_dir, exe_name, sources, console, unsigned, target_dir):
    file_name = os.path.join(output_dir, 'Makefile')
    sys.stdout.write("Generating %s'\n" % file_name)    
    f = open(file_name, 'w')
    f.write('#%s#\n' % ('=' * 76))
    f.write('# Makefile for %s %s #\n' % (exe_name, (' ' * (60 - len(exe_name)))))
    f.write('# This is a generated file; Do not modify it!                                #\n')
    f.write('#%s#\n' % ('-' * 76))

    # The standard python directories.
    f.write('PYINCDIR = "%s"\n' % os.path.join(sys.prefix, 'Include'))
    f.write('PYLIBDIR = "%s"\n' % os.path.join(sys.exec_prefix, 'libs'))
    f.write('PYLIBVER = %s%s\n' % (sys.version[0], sys.version[2]))

    # It seems that the required system libraries are linked in automatically
    # syslibs = ['kernel32', 'user32', 'gdi32', 'winspool', 'wsock32', 'comdlg32',
    #            'advapi32', 'shell32', 'ole32', 'oleaut32', 'odbc32', 'odbccp32']
    syslibs = []

    f.write('CC = cl /nologo\n')
    f.write('CXX = cl /nologo\n')
    f.write('LINK = link /nologo\n')
    f.write('RC = rc\n')
    f.write('CPPFLAGS = /DNDEBUG /I$(PYINCDIR)\n')
    f.write('CFLAGS = /Ox /MD /W3 /GX\n')
    f.write('CXXFLAGS = $(CFLAGS)\n')
    f.write('RFLAGS =\n')
    if console:
        f.write('LFLAGS = /MACHINE:IX86 /INCREMENTAL:NO /PDB:NONE /SUBSYSTEM:CONSOLE\n')
    else:
        f.write('LFLAGS = /MACHINE:IX86 /INCREMENTAL:NO /PDB:NONE /SUBSYSTEM:WINDOWS\n')

    f.write('SYSLIBS = %s\n\n' % string.join([x + '.lib' for x in syslibs]))

    if not unsigned:
        f.write('LIBS = $(PYLIBDIR)\\python$(PYLIBVER).lib $(PYLIBDIR)\\vendorid.lib\n\n')
    else:
        f.write('LIBS = $(PYLIBDIR)\\python$(PYLIBVER).lib\n\n')

    # Generate and write the list of object files
    objs = []
    for source in sources:
        ext = os.path.splitext(source)[1]
        if ext in ('.c', '.cc', '.cpp', '.cxx'):
            objs.append(os.path.splitext(os.path.basename(source))[0] + '.obj')
        elif ext == '.rc':
            objs.append(os.path.splitext(os.path.basename(source))[0] + '.res')
    prefix = 'FROZEN_OBJS = '
    for obj in objs:
        f.write('%s%s' % (prefix, obj))
        prefix = ' \\\n              '
    f.write('\n')

    # The target directory where to install
    f.write('TARGET_DIR = %s\n\n' % target_dir)

    # The target we want to create
    f.write('TARGET = %s.exe\n\n' % exe_name)
        
    f.write('all: $(TARGET)\n\n')

    f.write('$(TARGET): $(FROZEN_OBJS)\n')
    f.write('\t@if exist $(TARGET) del $(TARGET)\n')
    f.write('\t$(LINK) $(LFLAGS) /out:$(TARGET) @<<\n\t$(FROZEN_OBJS) $(LIBS) $(SYSLIBS)\n<<\n\n')

    # Generate the install target
    f.write('install: $(TARGET)\n')
    f.write('\t@if not exist $(TARGET_DIR) mkdir $(TARGET_DIR)\n')
    f.write('\tcopy $(TARGET) $(TARGET_DIR)\\$(TARGET)\n\n')

    # Generate the clean target
    f.write('clean:\n')
    f.write('\t-del $(TARGET)\n')
    f.write('\t-del *.res\n')
    f.write('\t-del *.obj\n\n')

    # Generic rule for rc sources
    f.write('{.}.rc{}.res:\n')
    f.write('\t$(RC) $(RFLAGS) $<\n\n')

    # Generic rule for C sources
    f.write('{.}.c{}.obj::\n')
    f.write('\t$(CC) -c $(CFLAGS) $(CPPFLAGS) -Fo @<<\n\t$<\n<<\n\n')
                
    # Generic rules for C++ sources
    f.write('{.}.cc{}.obj::\n')
    f.write('\t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -Fo @<<\n\t$<\n<<\n\n')
    f.write('{.}.cpp{}.obj::\n')
    f.write('\t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -Fo @<<\n\t$<\n<<\n\n')
    f.write('{.}.cxx{}.obj::\n')
    f.write('\t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -Fo @<<\n\t$<\n<<\n\n')

    f.write('#%s#\n' % ('=' * 76))
    f.write('# EOF%s#\n' % (' ' * 72))
    f.write('#%s#\n' % ('-' * 76))
    f.close()

#============================================================================#
# Generate a UNIX Makefile                                                   #
#----------------------------------------------------------------------------#
def mingw32_makefile(output_dir, exe_name, sources, console, unsigned, target_dir):
    file_name = os.path.join(output_dir, 'Makefile')
    sys.stdout.write("Generating %s'\n" % file_name)
    f = open(file_name, 'w')
    f.write('#%s#\n' % ('=' * 76))
    f.write('# Makefile for %s %s #\n' % (exe_name, (' ' * (60 - len(exe_name)))))
    f.write('# This is a generated file; Do not modify it!                                #\n')
    f.write('#%s#\n' % ('-' * 76))

    # The standard python directories.
    f.write('PYINCDIR = %s\n' % os.path.join(sys.prefix, 'include'))
    #f.write('PYLIBDIR = %s\n' % os.path.join(sys.exec_prefix, 'lib'))
    f.write('PYLIBSDIR = %s\n' % os.path.join(sys.exec_prefix, 'libs'))
    f.write('VERSION = %s%s\n' % (sys.version[0], sys.version[2]))

    ldlast = " -L$(PYLIBSDIR) " #-L$(PYLIBDIR)"

    # A signed interpreter needs the vendorid library.
    if not unsigned:        
        ldlast += ' -lvendorid'

    # Append the python library to the LDLAST variable.
    ldlast += " -lpython$(VERSION)"

    if not console:
        ldlast += ' -mwindows'
    
    # Modify the cpp flags
    cppflags = '-I. -I$(PYINCDIR)'
    f.write("CPPFLAGS = %s\n" % cppflags)
    f.write("LDLAST = %s\n" % ldlast)
    f.write("CC = gcc\n")
    f.write("CXX = g++\n")
    f.write("LINKCC = gcc\n")
    f.write('RC = windres\n')
    f.write("LIBS = \n")
    f.write("SYSLIBS = \n")

    # Generate and write the list of object files
    objs = []
    for source in sources:
        ext = os.path.splitext(source)[1]
        if ext in ('.c', '.cc', '.cpp', '.cxx'):
            objs.append(os.path.splitext(os.path.basename(source))[0] + '.o')
        elif ext == '.rc':
            objs.append(os.path.splitext(os.path.basename(source))[0] + '.o')
    prefix = 'FROZEN_OBJS = '
    for obj in objs:
        f.write('%s%s' % (prefix, obj))
        prefix = ' \\\n              '
    f.write('\n')

    # The target directory where to install
    f.write('TARGET_DIR = %s\n\n' % target_dir)

    # The target we want to create
    f.write('TARGET = %s.exe\n\n' % exe_name)
        
    # Extend the list of suffixes
    f.write(".SUFFIXES: .rc\n\n")

    f.write('all: $(TARGET)\n\n')

    f.write('$(TARGET): $(FROZEN_OBJS)\n')
    f.write('\t$(LINKCC) $(LDFLAGS) $(LINKFORSHARED) $(FROZEN_OBJS) -o $(TARGET) $(LIBS) $(SYSLIBS) $(LDLAST)\n\n')


    # Generate the install target
    f.write('install: $(TARGET)\n')
    f.write('\t@if not exist $(TARGET_DIR) mkdir $(TARGET_DIR)\n')
    f.write('\tcopy $(TARGET) $(TARGET_DIR)\\$(TARGET)\n\n')

    # Generate the clean target
    f.write('clean:\n')
    f.write('\t-del $(TARGET)\n')
    f.write('\t-del *.o\n\n')

    # Generic rule for rc sources
    f.write('.rc.o:\n')
    f.write('\t$(RC) $(RFLAGS) -i $< -o $@\n\n')

    # Generic rule for C sources
    f.write('.c.o:\n')
    f.write('\t$(CC) $(CFLAGS) $(CPPFLAGS) -c  $< -o $@\n\n')

    # Generic rule for C++ sources
    f.write('.cc.o:\n')
    f.write('\t$(CXX) $(CFLAGS) $(CPPFLAGS) -c  $< -o $@\n\n')                
    f.write('.cpp.o:\n')
    f.write('\t$(CXX) $(CFLAGS) $(CPPFLAGS) -c  $< -o $@\n\n')
    f.write('.cxx.o:\n')
    f.write('\t$(CXX) $(CFLAGS) $(CPPFLAGS) -c  $< -o $@\n\n')
                
    f.write('#%s#\n' % ('=' * 76))
    f.write('# EOF%s#\n' % (' ' * 72))
    f.write('#%s#\n' % ('-' * 76))
    f.close()


#============================================================================#
# Usage information                                                          #
#----------------------------------------------------------------------------#
def usage(name):
    sys.stdout.write("\nUsage: %s [Options] main-script\n\n" % name)
    sys.stdout.write("This script takes a python main script and a number of packages and modules\n")
    sys.stdout.write("and generates C sources and a Makefile. After calling 'make' in the build\n")
    sys.stdout.write("directory, you should get an executable containing the main script and all\n")
    sys.stdout.write("additional packages and modules.\n\n")
    sys.stdout.write("Options: -n <name>     The name of the executable. The default is based on the\n")
    sys.stdout.write("                       name of the main script.\n")
    sys.stdout.write("         -u            Do not create a signed interpreter.\n")
    sys.stdout.write("         -v            Add support for Python's verbose flag.\n")
    sys.stdout.write("         -o #          Optimize flag.\n")
    sys.stdout.write("                       0 = no optimizing (the default).\n")
    sys.stdout.write("                       1 = normal optimizing (like python -O).\n")
    sys.stdout.write("                       2 = additionally removes doc strings (like python -OO).\n")
    sys.stdout.write("         -p <prefix>   Use <prefix> as path prefix for compiled python code.\n")
    sys.stdout.write("                       Default is the absolute path of the python source file.\n")
    sys.stdout.write("         -d <dir>      Use <dir> for the generated output files. Default is\n")
    sys.stdout.write("                       'build_' + the base name of the python main script.\n")
    sys.stdout.write("         -m <package>  <package> is a package or module that should be included\n")
    sys.stdout.write("                       into the executable. Use this option for every top level\n")
    sys.stdout.write("                       package or module that should be part of the executable.\n")
    sys.stdout.write("                       Subpackages of a package are included automatically.\n")
    sys.stdout.write("         -t <dir>      Install the application in <dir> when you call\n")
    sys.stdout.write("                       'make install' for the generated Makefile. Default is\n")
    sys.stdout.write("                       to install it into the same directory where the python\n")
    sys.stdout.write("                       binary is installed.\n")
    if sys.platform == 'win32':
        sys.stdout.write("         -c            Create a console application. Default is to build a\n")
        sys.stdout.write("                       windows application.\n")
        sys.stdout.write("         -i <file.ico> Use <file.ico> as the applications icon file. This\n")
        sys.stdout.write("                       must be a valid ico file. If this option is not used,\n")
        sys.stdout.write("                       a builtin default icon is used.\n")
        sys.stdout.write("         -M            Use MinGW32 instead of MSVC (default) for compilation.")

#============================================================================#
# ----- Main Routine ------------------------------------------------------- #
#----------------------------------------------------------------------------#
if __name__ == '__main__':
    # Parse and process the call arguments
    executable = os.path.basename(sys.argv[0])
    flags = 'n:m:d:t:p:uvo:'
    if sys.platform == 'win32':
        flags += 'ci:M'
    try:
        optlist, args = getopt.getopt(sys.argv[1:], flags)
    except getopt.GetoptError:
        usage(executable)
        sys.exit(1)
    if not args:
        usage(executable)
        sys.exit(1)
    if len(args) != 1:
        usage(executable)
        sys.exit(1)
    main_script = args[0]
    exe_name = os.path.splitext(os.path.basename(main_script))[0]
    output_dir = 'build_' + exe_name
    if sys.platform == 'win32':
        target_dir = sys.exec_prefix
    else:
        target_dir = os.path.join(sys.exec_prefix, 'bin')
    icon_file = None
    prefix = None
    module_list = []
    optimize = 0
    verbose = 0
    console = 0
    unsigned = 0
    use_mingw32 = 0
    for opt, arg in optlist:
        if opt == "-n":
            exe_name = arg
        elif opt == "-m":
            module_list.append(arg)
        elif opt == "-d":
            output_dir = arg
        elif opt == "-p":
            prefix = arg
        elif opt == "-u":
            unsigned = 1
        elif opt == "-v":
            verbose = 1
        elif opt == "-o":
            try:
                optimize = int(arg)
            except:
                usage(executable)
        elif opt == "-t":
            target_dir = arg
        elif sys.platform == 'win32' and opt == "-c":
            console = 1
        elif sys.platform == 'win32' and opt == "-i":
            icon_file = arg
        elif sys.platform == 'win32' and opt == "-M":
            use_mingw32 = 1

    # Check the main script
    if not os.path.exists(main_script):
        sys.stderr.write("The main script '%s' does not exist\n" % main_script)
        sys.exit(1)
    if not os.path.isfile(main_script):
        sys.stderr.write("The main script '%s' is not a regular file\n" % main_script)
        sys.exit(1)
    if os.path.splitext(main_script)[1] != '.py':
        sys.stderr.write("The main script '%s' has no .py extension\n" % main_script)
        sys.exit(1)

    # Check the modules/packages
    for name in module_list:
        if not os.path.exists(name):
            sys.stderr.write("The package or module '%s' does not exist\n" % name)
            sys.exit(1)
        if os.path.isdir(name) and not is_package(name):
            sys.stderr.write("The package '%s' is not a valid python package\n", name)
            sys.exit(1)
        if os.path.isfile(name) and os.path.splitext(name)[1] != '.py':
            sys.stderr.write("The module '%s' is not a valid python module\n" % name)
            sys.exit(1)

    # Check the icon file
    if sys.platform == 'win32' and icon_file:
        if not os.path.exists(icon_file):
            sys.stderr.write("The icon file '%s' does not exist\n" % icon_file)
            sys.exit(1)
        if not os.path.isfile(icon_file):
            sys.stderr.write("The icon file '%s' is not a regular file\n" % icon_file)
            sys.exit(1)
            
    # Check the value of the optimize flag
    if optimize not in (0, 1, 2):
        sys.stderr.write("The value of the '-o' flag must be 0, 1 or 2\n")
        sys.exit(1)

    # Freeze the main script
    module_name = '__main__'
    if prefix is None:
        pfx = os.path.abspath(main_script)
    else:
        pfx = os.path.join(prefix, os.path.basename(main_script))
    frozen_main = FrozenCode(module_name, main_script, pfx)

    # This list contains all generated source files
    sources = []

    # This list contains all frozen packages and modules
    frozen_modules = []

    # Freeze all modules/packages
    for name in module_list:
        if is_package(name):
            resolve_package(name, '', prefix, frozen_modules)
        else:
            module_name = os.path.splitext(os.path.basename(name))[0]
            if prefix is None:
                p = os.path.abspath(name)
            else:
                p = os.path.join(prefix, os.path.basename(name))
            frozen_modules.append(FrozenCode(module_name, name, p))

    # If the build directory does not already exist, create it
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, 0755)
        
    # Write the C source of the main script
    file_name = os.path.join(output_dir, frozen_main.mangled_name + '.c')
    sys.stdout.write("Generating main script code '%s'\n" % file_name)
    f = open(file_name, 'w')
    f.write('/*%s*/\n' % ('=' * 74))
    f.write('/* Frozen main script for %s %s*/\n' % (exe_name, (' ' * (49 - len(exe_name)))))
    f.write('/* Generated from %s %s*/\n' % (frozen_main.input_file, (' ' * (57 - len(frozen_main.input_file)))))
    f.write('/* This is generated code; Do not modify it!                                */\n')
    f.write('/*%s*/\n' % ('-' * 74))
    f.write(frozen_main.c_code)
    f.write('/*%s*/\n' % ('=' * 74))
    f.write('/* EOF%s*/\n' % (' ' * 70))
    f.write('/*%s*/\n' % ('-' * 74))
    f.close()
    sources.append(file_name)

    # Write the C sources of the packages and modules
    for fm in frozen_modules:
        file_name = os.path.join(output_dir, fm.mangled_name + '.c')
        sys.stdout.write("Generating module code '%s'\n" % file_name)    
        f = open(file_name, 'w')
        f.write('/*%s*/\n' % ('=' * 74))
        f.write('/* Frozen module code for %s %s*/\n' % (exe_name, (' ' * (49 - len(exe_name)))))
        f.write('/* Generated from %s %s*/\n' % (fm.input_file, (' ' * (57 - len(fm.input_file)))))
        f.write('/*%s*/\n' % ('-' * 74))
        f.write(fm.c_code)
        f.write('/*%s*/\n' % ('=' * 74))
        f.write('/* EOF%s*/\n' % (' ' * 70))
        f.write('/*%s*/\n' % ('-' * 74))
        f.close()
        sources.append(file_name)

    # On UNIX, parse the Python Makefile and check, if the main code must be C++
    use_cxx = 0
    if sys.platform != 'win32':
        makevars = parse_unix_makefile()
        if makevars.has_key('MAINOBJ') and makevars['MAINOBJ'] == 'ccpython.o':
            use_cxx = 1
        
    # Generate the main C code
    generate_main(output_dir, exe_name, frozen_main, frozen_modules,
                  console, unsigned, optimize, verbose, sources, use_cxx)

    if sys.platform == 'win32':
        # Generate the recource file
        generate_rc(output_dir, exe_name, icon_file, console, sources)
        # Generate the Makefile
        if use_mingw32:
            mingw32_makefile(output_dir, exe_name, sources, console, unsigned, target_dir)
        else:
            win32_makefile(output_dir, exe_name, sources, console, unsigned, target_dir)
    else:
        # Generate the Makefile
        unix_makefile(output_dir, exe_name, sources, makevars, unsigned, target_dir)
        
#============================================================================#
# EOF                                                                        #
#----------------------------------------------------------------------------#